from random import shuffle, choice
import libtcodpy as tcod

class GenerationError(Exception):
    pass

def generate(width, height):
    def generate_walls():
        cells = dict(((x,y),       (x,y)) for x in xrange(width)   for y in xrange(height))
        walls = dict(((x,y,True),  None)  for x in xrange(width-1) for y in xrange(height))
        walls.update(((x,y,False), None)  for x in xrange(width)   for y in xrange(height-1))
        keys = walls.keys()
        shuffle(keys)
        
        def root(cell):
            parent = cells[cell]
            if parent != cell:
                cells[cell] = parent = root(parent)
            return parent
        
        for key in keys:
            x, y, vertical = key
            if vertical:
                other = (x+1,y)
            else:
                other = (x,y+1)
            parent = root(other)
            main = root((x,y))
            if parent != main:
                walls[key] = choice("WUBRG+++++")
                cells[parent] = main
        
        return walls

    def directions(x,y):
        if x:
            yield (x-1, y, True),  (x-1, y)
        if y:
            yield (x, y-1, False), (x, y-1)
        if x < width-1:
            yield (x, y, True),    (x+1, y)
        if y < height-1:
            yield (x, y, False),   (x, y+1)

    def generate_ends():
        cells = dict(((x,y), "+") for x in xrange(width) for y in xrange(height))
        cell = 0,0
        last = None
        
        # Find a start cell
        while True:
            neighbors = [c for wall,c in directions(*cell) if c != last and walls[wall]]
            if not neighbors:
                break
            last = cell
            cell = choice(neighbors)
        cells[cell] = "@"
        for wall, other in directions(*cell):
            if walls[wall]:
                walls[wall] = "+"
        
        # Find a goal cell
        seen = set()
        boundary = [cell]
        while boundary:
            last = boundary
            seen.update(boundary)
            boundary = [
                c
                for neighbor in boundary
                for wall,c in directions(*neighbor)
                if c not in seen and walls[wall]
            ]
        goal = choice(last)
        cells[goal] = "$"
        
        return cells, cell, goal
    
    def fromcell(cell, wall):
        seen = set()
        boundary = [cell]
        while boundary:
            seen.update(boundary)
            boundary = [
                c
                for neighbor in boundary
                for w,c in directions(*neighbor)
                if c not in seen and w != wall and walls[w]
            ]
        return seen
    
    # Don't allow allies to be behind each other.
    ally = choice([1,4])
    def passable(color):
        colors = "WUBRG"
        c = colors.find(color)
        allied = colors[(c+ally)%5]
        return set([allied, "+"])
    
    def expand(cell, allowed):
        seen = set()
        boundary = [cell]
        while boundary:
            seen.update(boundary)
            boundary = [
                c
                for neighbor in boundary
                for w,c in directions(*neighbor)
                if c not in seen and walls[w] in allowed
            ]
        return seen
    
    def enemies(color):
        colors = "wubrg"
        c = colors.find(color)
        return set(colors[(c+i)%5] for i in (2,3))
    
    def generate_switches():
        trim = {}
        for wall in walls:
            color = walls[wall]
            if color and color != "+":
                x, y, vertical = wall
                if vertical:
                    sides = (x,y), (x+1,y)
                else:
                    sides = (x,y), (x,y+1)
                
                for cell in sides:
                    required = cell in fromcell(start, wall)
                    if not required:
                        if cell in fromcell(goal, wall):
                            # No need to go back.
                            continue
                    
                    possible = expand(cell, passable(color))
                    switches = set(cells[c] for c in possible)
                    switch = color.lower()
                    if switch in switches:
                        # An accessible switch already exists.
                        continue
                    
                    empty = [c for c in possible if cells[c] == "+"]
                    if empty:
                        cell = choice(empty)
                        cells[cell] = switch
                    else:
                        raise GenerationError("Unable to find a site!")
                    
                    if not switches & enemies(switch):
                        trim[cell] = wall
        
        # Trim any useless switches
        while trim:
            trimmed = False
            untrimmed = {}
            for cell in trim:
                wall = trim[cell]
                switch = cells[cell]
                allowed = set("WUBRG+")
                allowed.discard(switch.upper())
                influence = expand(cell, allowed)
                if start in influence:
                    # Don't trim the first switch.
                    continue
                
                switches = set(cells[c] for c in influence)
                if switches & enemies(switch):
                    # Still a way to trap yourself.
                    untrimmed[cell] = wall
                else:
                    #print "Trimming %s: %s" % (cell, switch)
                    cells[cell] = "+"
                    trimmed = True
            if trimmed:
                trim = untrimmed
            else:
                break

    def wall(blocked):
        return blocked or " "
    
    while True:
        try:
            walls = generate_walls()
            cells, start, goal = generate_ends()
            generate_switches()
        except GenerationError:
            print "Level generation failed; retrying..."
        else:
            break
    
    for y in xrange(height-1):
        line = []
        for x in xrange(width-1):
            line.append(cells[x,y])
            line.append(wall(walls[x,y,True]))
        line.append(cells[width-1,y])
        yield line
        
        line = []
        for x in xrange(width):
            line.append(wall(walls[x,y,False]))
            line.append(" ")
        yield line
    
    y = height - 1
    line = []
    for x in xrange(width-1):
        line.append(cells[x,y])
        line.append(wall(walls[x,y,True]))
    line.append(cells[width-1,y])
    yield line

def display():
    w, h = 40, 20
    for line in generate(w, h):
        print str.join("", line)
    
    tcod.console_init_root(w*3, h*3, "Map Generation", False)
    color = tcod.grey
    for y, line in enumerate(generate(w, h)):
        for x, char in enumerate(line):
            if char == "+":
                tcod.console_set_char_background(0, x+2,y+2, color)
    tcod.console_flush()
    tcod.console_wait_for_keypress(True)

if __name__ == "__main__":
    display()
