from collections import defaultdict
from rainbowrooms.world import Rock, Tile, Wall
import libtcodpy as tcod
import logging

class Window(object):
    def __init__(self):
        # Basic libtcod initialization
        h, w = 80, 150
        self.width = w
        self.height = h
        tcod.console_init_root(w, h, "Rainbow Rooms", False)
        
        # Logging
        logging.basicConfig(level=logging.DEBUG)
        self.logger = logging.getLogger("rainbowrooms")
        
        # Subwindows
        mapcols = w - 20
        maplines = h - 4
        self.map = tcod.console_new(mapcols, maplines)
        self.view = tcod.map_new(mapcols, maplines)
        self.msg = None
        
        # Key names
        self.keynames = {}
        for name in dir(tcod):
            if name.startswith("KEY_"):
                self.keynames[getattr(tcod, name)] = name[4:]
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        # Clean up the windows
        pass
    
    def display(self, world):
        tcod.console_clear(self.map)
        px, py, pz = world.player.loc
        
        h = tcod.console_get_height(self.map)
        w = tcod.console_get_width(self.map)
        assert 0 <= px < w
        assert 0 <= py < h
        radius = 0
        ploc = (px, py)
        
        seen = {}
        ramped = set(pz)
        view = self.view
        tcod.map_clear(view)
        neighbors = defaultdict(set)
        while ramped:
            # Add new spaces to the unseen space.
            z = ramped.pop()
            level = world.level(z)
            ramps = {}
            for loc in level.spaces:
                if loc not in seen:
                    x, y = loc
                    space = level.spaces[loc]
                    transparent = space.transparent
                    if transparent and space.passable and space.passable != (z,):
                        # We might see another level on the other side.
                        ramps[loc] = set(dz for dz in space.passable if dz != z)
                    tcod.map_set_properties(view, x, y, space.transparent, True)
                elif loc in neighbors[z]:
                    # Override the Rock previously seen through a ramp.
                    x, y = loc
                    space = level.spaces[loc]
                    if seen[loc] not in (Rock, space):
                        print "Replacing %r with %r at %s" % (seen[loc], space, loc)
                    seen[loc] = space
                    tcod.map_set_properties(view, x, y, space.transparent, True)
            
            # Re-compute the field of view.
            tcod.map_compute_fov(view, px,py, radius, True, tcod.FOV_PERMISSIVE_3)
            
            for x in xrange(w):
                for y in xrange(h):
                    loc = x,y
                    if loc not in seen:
                        if tcod.map_is_in_fov(view, x, y):
                            # Collect the newly seen space.
                            space = level.spaces.get(loc, Rock)
                            seen[loc] = space
                            if loc in ramps:
                                others = ramps[loc]
                                ramped.update(others)
                                nx = [0]
                                if px <= x: nx.append(1)
                                if px >= x: nx.append(-1)
                                ny = [0]
                                if py <= y: ny.append(1)
                                if py >= y: ny.append(-1)
                                neighborhood = set((x+dx, y+dy) for dx in nx for dy in ny)
                                neighborhood.discard((x,y))
                                for oz in others:
                                    neighbors[oz].update(neighborhood)
                        else:
                            # Ignore this space on other levels.
                            tcod.map_set_properties(view, x, y, False, False)
        
        for loc in seen:
            space = seen[loc]
            self.cell(loc, space.symbol, space.color, Tile.color)
        self.cell(ploc, tcod.CHAR_SMILIE, tcod.lightest_grey)
        
        tcod.console_blit(self.map, 0,0, 0,0, 0, 2,2)
        
        # Guide to each color's state.
        locs = [(0,2), (6,0), (10,5), (6,10), (0,8)]
        pixels = [
            [0x80, 0x20, 0x90],
            [0x20, 0x20, 0x20],
            [0x81, 0x20, 0x91],
        ]
        
        for n, (x,y) in enumerate(locs):
            color = Wall.colors[n][Wall.switched[n]]
            for dy, row in enumerate(pixels):
                for dx, char in enumerate(row):
                    tcod.console_put_char_ex(0, x+w+2+dx, y+2+dy, char, tcod.black, color)
        
        if self.msg:
            self.display_message()
        
        tcod.console_flush()
    
    def display_message(self):
        width = self.width >> 1
        textwidth = width - 4
        height = 4 + tcod.console_get_height_rect(0, 0, 0, textwidth, self.height, self.msg)
        window = tcod.console_new(width, height)
        tcod.console_set_default_foreground(window, tcod.white)
        tcod.console_print_frame(window, 0, 0, width, height, False, tcod.BKGND_NONE)
        tcod.console_print_rect_ex(window, width >> 1, 2, textwidth, height, tcod.BKGND_NONE, tcod.CENTER, self.msg)
        
        #blit the contents of "window" to the root console
        x = (self.width>>1) - (width>>1)
        y = (self.height>>1) - (height>>1)
        tcod.console_blit(window, 0, 0, width, height, 0, x, y, 1.0, 0.7)
    
    def cell(self, loc, symbol, color=None, bg=None):
        try:
            # Map strings to numbers.
            symbol = ord(symbol)
        except TypeError:
            pass
        
        if not symbol:
            symbol = " "
            bg, color = color, bg
        
        x, y = loc
        assert 0 <= x < tcod.console_get_width(self.map)
        assert 0 <= y < tcod.console_get_height(self.map)
        if color: tcod.console_set_char_foreground(self.map, x,y, color)
        if bg: tcod.console_set_char_background(self.map, x,y, bg)
        tcod.console_set_char(self.map, x,y, symbol)
    
    def event(self):
        self.msg = None
        while True:
            try:
                char = tcod.console_wait_for_keypress(False)
                if not char.pressed:
                    # Ignore key up events.
                    # Accepting key down allows us to use repeats,
                    # but accepting both up and down duplicates commands.
                    continue
                
                if char.vk == tcod.KEY_CHAR:
                    key = chr(char.c)
                else:
                    key = self.keynames.get(char.vk, str(char.vk))
                    if char.shift:
                        key = "S-" + key
                
                if char.lctrl or char.rctrl:
                    key = "^" + key
            except KeyboardInterrupt:
                key = "^c"
            
            return key
    
    def message(self, msg):
        self.msg = msg
    
    def log(self, message, *args):
        if args:
            format = message
        else:
            format = "%s"
            args = message,
        self.logger.info(format, *args)
