from random import randrange
from rainbowrooms.player import Character
from rainbowrooms.data import load
from rainbowrooms.mapgen import generate
import libtcodpy as tcod

def neighborhood(x,y):
    result = set(
        (x+dx, y+dy)
        for dx in (-1, 0, 1)
        for dy in (-1, 0, 1)
    )
    
    result.discard((x,y))
    return result

class Tile(object):
    symbol = None
    passable = False
    transparent = True
    color = tcod.darkest_grey
    
    @classmethod
    def bump(self):
        pass

class Space(Tile):
    colors = tcod.color_gen_map([tcod.darkest_grey, tcod.dark_grey], [0, 10])
    
    def __init__(self, levels, has_player=False):
        self.passable = levels
        self.travelled = has_player
    
    def bump(self):
        if self.travelled < len(self.colors) - 1:
            self.travelled += 1
    
    @property
    def color(self):
        return self.colors[self.travelled]

class Wall(Tile):
    switched = [True] * 5
    colors = [
        (tcod.darkest_flame,   tcod.flame),
        (tcod.darkest_yellow,  tcod.yellow),
        (tcod.darkest_sea,     tcod.sea),
        (tcod.darkest_azure,   tcod.azure),
        (tcod.darkest_fuchsia, tcod.fuchsia),
    ]
    
    def __init__(self, levels, index):
        self.index = index
        self.z = levels
    
    @property
    def color(self):
        lit = self.switched[self.index]
        return self.colors[self.index][lit]
    
    @property
    def passable(self):
        if self.switched[self.index]:
            return False
        else:
            return self.z
    
    @property
    def transparent(self):
        return not self.switched[self.index]

class Switch(Wall):
    symbol = tcod.CHAR_BULLET
    passable = False
    transparent = True
    colors = [
        (tcod.darker_flame,   tcod.flame),
        (tcod.darker_yellow,  tcod.yellow),
        (tcod.darker_sea,     tcod.sea),
        (tcod.darker_azure,   tcod.azure),
        (tcod.darker_fuchsia, tcod.fuchsia),
    ]
    
    def bump(self):
        self.switched[self.index] = not self.switched[self.index]
        self.switched[(self.index + 2) % 5] = True
        self.switched[(self.index + 3) % 5] = True

class Goal(Tile):
    symbol = tcod.CHAR_HEART
    color = tcod.red
    
    def __init__(self, levels):
        self.passable = levels
    
    def bump(self):
        self.symbol = tcod.CHAR_LIGHT
        self.color = tcod.yellow
        return True

class Rock(Tile):
    symbol = tcod.CHAR_BLOCK2
    passable = False
    transparent = False
    color = tcod.sepia

class Level(object):
    def __init__(self, size):
        self.spaces = {}
        self.size = size
    
    def load(self, z, previous):
        loc = (2, 2)
        size, depth = self.size
        world = generate(size + 4, size + 2)
        
        self.z = (z,)
        self.down = (z, z+1)
        self.above = previous
        self.static = {}
        self.static["$"] = Goal(self.z)
        for index, char in enumerate("WUBRG"):
            self.static[char] = Wall(self.z, index)
            self.static[char.lower()] = Switch(self.z, index)
        
        handlers = [
            [self.room, self.hall_horiz],
            [self.hall_vert, self.ignore],
        ]
        
        print "Loading level " + str(z)
        for y, line in enumerate(world):
            for x, char in enumerate(line):
                player = handlers[y%2][x%2](x, y, char)
                if player:
                    loc = player
        
        return loc
    
    def room(self, x, y, char):
        x = 4*(x>>1) + 2
        y = 4*(y>>1) + 2
        
        loc = None
        tile = None
        if char == "@":
            loc = (x, y)
            tile = Space(self.z, 1)
        elif char == "+":
            tile = Space(self.z)
        elif char in self.static:
            tile = self.static.get(char)
        else:
            # Empty room.
            return
        
        self.spaces[x,y] = tile
        for dx,dy in neighborhood(x,y):
            self.spaces[dx,dy] = Space(self.z)
        
        return loc
    
    def hall_horiz(self, x, y, char):
        x = 4*(x>>1) + 4
        y = 4*(y>>1) + 1
        self.hallway(x, y, char, [0], [0, 1, 2])
    
    def hall_vert(self, x, y, char):
        x = 4*(x>>1) + 1
        y = 4*(y>>1) + 4
        self.hallway(x, y, char, [0, 1, 2], [0])
    
    def hallway(self, x, y, char, dxs, dys):
        tile = None
        if char == "+":
            levels = self.z
        elif char == ">":
            levels = self.down
        elif char == "<":
            z = self.z[0]
            up = (z-1, z)
            for dx in dxs:
                for dy in dys:
                    self.spaces[x+dx,y+dy] = tile = self.above.spaces[x+dx,y+dy]
                    assert tile.passable == up
            return
        else:
            tile = self.static.get(char)
            if not tile:
                return
        
        for dx in dxs:
            for dy in dys:
                self.spaces[x+dx,y+dy] = tile or Space(levels)
    
    def ignore(self, x, y, char):
        pass

class World(object):
    maxsize = 12
    
    def __init__(self):
        self.size = [0,0]
        self.reset()
    
    def reset(self):
        self.size[randrange(self.maxsize) < self.size[0]] += 1
        level = Level(self.size)
        x, y = level.load(0, None)
        z = (0,)
        self.levels = [level]
        self.player = Character((x, y, z))
    
    def __getitem__(self, loc):
        x, y, levels = loc
        for z in levels:
            space = self.level(z).spaces.get((x,y))
            if space:
                return space
        return Rock
    
    def level(self, z):
        maxlev = len(self.levels)
        assert 0 <= z <= maxlev
        if z == maxlev:
            level = Level(self.size)
            level.load(z, self.levels[-1])
            self.levels.append(level)
        else:
            level = self.levels[z]
        return level
